package neuralNetwork;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Random;

public class MLPLMT21 {
	public int layerN;
	public int[] unitN;
	public double[][][] unit;
	public double[][][][] w;
	double[][][] delta;
	double[][][] deltaCache;

	double[][][][] wDelta;
	int outLayer;
	double err;
	int numGamete	= 4;
	double[] errGamete;
	double[] errGameteOld;
	double[] evalFuncValue;
	double mutationRateDominant = 0.0;
	double mutationRateRecessive = 0.0;
//	Random rand = new Random();
	Random rand = new Random(0);

	public MLPLMT21(int layerN, int[] unitN,double wMax) {
		this.layerN = layerN;
		this.unitN = unitN;
		outLayer = layerN - 1;
		unit = new double[numGamete][layerN][];
		delta = new double[numGamete][layerN][];
		deltaCache = new double[numGamete][layerN][];	// layerN, unitN
		w = new double[numGamete][layerN][][];
		wDelta = new double[numGamete][layerN][][];
		evalFuncValue = new double[numGamete];
		errGamete = new double[numGamete];
		errGameteOld = new double[numGamete];

		for (int k=0; k<numGamete; k++){
			for (int l = 0; l < layerN; l++) {
				int u = unitN[l];
				unit[k][l] = new double[u + (l == outLayer ? 0 : 1)];
				if (l < outLayer) {
					unit[k][l][u] = 1.0;
					delta[k][l] = new double[u];
					deltaCache[k][l] = new double[u];
				}
				if (l > 0) {
					int v = unitN[l-1]+1;
					delta[k][l] = new double[u];
					deltaCache[k][l] = new double[u];
					w[k][l] = new double[u][v];
					wDelta[k][l] = new double[u][v];
				}
			}
			for (int l = 1; l < layerN; l++) {
				for (int j = 0; j < unitN[l]; j++) {
					for (int i = 0; i < unitN[l-1]+1; i++) {
						w[k][l][j][i] = (rand.nextDouble()*2 - 1) * wMax;
					}
				}
			}
		}
	}

	double sigmoid(double x) {
		 return 1 / (1 + Math.pow(Math.E, -x));
	}

	void softmax(double[] x) {
		if (x.length == 1) {
			for (int i = 0; i < x.length; i++) x[i] = sigmoid(x[i]);
		} else {
			double s = 0.0;
			for (int i = 0; i < x.length; i++) {
				x[i] = Math.exp(x[i]);
				s += x[i];
			}
			for (int i = 0; i < x.length; i++) x[i] /= s;
		}
	}

	public void forward(double[] d) {
		for (int k=0; k<numGamete; k++){
			for (int j = 0; j < unitN[0]; j++) unit[k][0][j] = d[j];
			for (int l = 0; l < outLayer-1; l++) {
				for (int j = 0; j < unitN[l+1]; j++) {
					double s = 0.0;
					for (int i = 0; i < unitN[l]+1; i++) {
						s += w[k][l+1][j][i] * unit[k][l][i];
					}
					unit[k][l+1][j] = sigmoid(s);
				}
			}

			for (int j = 0; j < unitN[outLayer]; j++) {
				unit[k][outLayer][j] = 0.0;
				for (int i = 0; i < unitN[outLayer-1]+1; i++) {
					unit[k][outLayer][j] +=
							w[k][outLayer][j][i] * unit[k][outLayer-1][i];
				}
			}
			softmax(unit[k][outLayer]);
		}
	}

	void backPropagate(double[] d, double[] t) {
		for (int m=0; m<numGamete; m++){
			err = errGamete[m];
			for (int j = 0; j < unitN[outLayer]; j++) {
				double e = t[j] - unit[m][outLayer][j];
				delta[m][outLayer][j] = e;
				err += e * e;
				evalFuncValue[m] += -(t[j]*Math.log(unit[m][outLayer][j]));
			}
			errGamete[m] = err;
			for (int l = outLayer-1; l > 0; l--) {
				for (int j = 0; j < unitN[l]; j++) {
					double df = unit[m][l][j] * (1.0 - unit[m][l][j]);
					double s = 0.0;
					for (int k = 0; k < unitN[l+1]; k++) {
						s+= delta[m][l+1][k] * w[m][l+1][k][j];
					}
					delta[m][l][j] += df * s;
				}
			}
		}
	}

	void recombination2(int testSwitch) {
		for (int k=0; k < numGamete; k++){
			for (int l = layerN-1; l > 0; l--) {
				double[][][] wTmp = new double[numGamete][unitN[l]][unitN[l-1]+1];
				int numRecombination;
				double recombRate = rand.nextGaussian()/2.0 + 1.0;
				double[] randFathertemp = new double[]{1.0, 1.0, 1.0, 1.0};
				double[] randMothertemp = new double[]{1.0, 1.0, 1.0, 1,0};

				if (recombRate < 0.5) numRecombination = 0;
				else if (recombRate < 1.5) numRecombination = 1;
				else if (recombRate < 2.5) numRecombination = 2;
				else if (recombRate < 3.5) numRecombination = 3;
				else numRecombination = 4;

				for (int randloop = 0; randloop< numRecombination; randloop++) {
					randFathertemp[randloop] = rand.nextDouble();
				}
				Arrays.sort(randFathertemp);

				for (int randloop = 0; randloop< numRecombination; randloop++) {
					randMothertemp[randloop] = rand.nextDouble();
				}
				Arrays.sort(randMothertemp);

				for (int j = 0; j < unitN[l]; j++) {
					for (int i = 0; i < unitN[l-1]+1; i++) {

						double tmpRatio = (double) i/ (double) (unitN[l-1]+1);
						int tmpRecombCount = 0;
						for (int ll = 0; ll < numRecombination; ll++) {
							if (tmpRatio > randFathertemp[ll]) tmpRecombCount = ll + 1;
						}
						int tmpRecombFlag = 0;
						if ((tmpRecombCount % 2) == 1) tmpRecombFlag = 1;
						if (tmpRecombFlag ==1)
							wTmp[0][j][i] = w[1][l][j][i];
						else
							wTmp[0][j][i] = w[0][l][j][i];

						tmpRecombCount = 0;
						for (int ll = 0; ll < numRecombination; ll++) {
							if (tmpRatio > randMothertemp[ll]) tmpRecombCount = ll + 1;
						}
						tmpRecombFlag = 0;
						if ((tmpRecombCount % 2) == 1) tmpRecombFlag = 1;
						if (tmpRecombFlag ==1)
							wTmp[1][j][i] = w[3][l][j][i];
						else
							wTmp[1][j][i] = w[2][l][j][i];

						wTmp[2][j][i] = (w[0][l][j][i] + w[1][l][j][i]) / 2.0;

						wTmp[3][j][i] = (w[2][l][j][i] + w[3][l][j][i]) / 2.0;
					}
				}
				w[0][l] = wTmp[0]; w[1][l] = wTmp[1]; w[2][l] = wTmp[2]; w[3][l] = wTmp[3];

			}
		}
	}

	void update21(int testSwitch, double rate, int batchSize) {
		int maxPointer1st = 0;		// max pointer in all Gametes
		int maxPointer2nd = 0;	// 2nd pointer between maxpointer in father and in mother
		int maxPointer1stOther = 0;		// max pointer in all Gametes
		int maxPointer2ndOther = 0;	// 2nd pointer between maxpointer in father and in mother
		int maxPointerFather1 = 0;	// max pointer in Father Gametes
		int maxPointerFather2 = 0;	// 2nd pointer in Father Gametes
		int maxPointerMother1 = 0;	// max pointer in Mother Gametes
		int maxPointerMother2 = 0;	// 2nd pointer in Mother Gametes

		for (int k=0; k < numGamete; k++){
			for (int l = layerN-1; l > 0; l--) {
				double[][][] wTmp = new double[numGamete][unitN[l]][unitN[l-1]+1];
				for (int j = 0; j < unitN[l]; j++) {
					for (int i = 0; i < unitN[l-1]+1; i++) {
						maxPointerFather1 = (evalFuncValue[0] <= evalFuncValue[1] ? 0 : 1);
						maxPointerFather2 = (evalFuncValue[0] > evalFuncValue[1] ? 0 : 1);
						maxPointerMother1 = (evalFuncValue[2] <= evalFuncValue[3] ? 2 : 3);
						maxPointerMother2 = (evalFuncValue[2] > evalFuncValue[3] ? 2 : 3);
						maxPointer1st =
								(evalFuncValue[maxPointerFather1]
										<= evalFuncValue[maxPointerMother1] ? maxPointerFather1 : maxPointerMother1);
						maxPointer1stOther =
								(maxPointer1st == maxPointerFather1 ? maxPointerFather2 : maxPointerMother2);
						maxPointer2nd =
								(evalFuncValue[maxPointerFather1]
										> evalFuncValue[maxPointerMother1] ? maxPointerFather1 : maxPointerMother1);
						maxPointer2ndOther =
								(maxPointer2nd == maxPointerFather1 ? maxPointerFather2 : maxPointerMother2);

						wTmp[0][j][i] = w[maxPointer1st][l][j][i];
						wTmp[0][j][i] += rate * delta[maxPointer1st][l][j]*unit[maxPointer1st][l-1][i];

						wTmp[1][j][i] = w[maxPointer2nd][l][j][i];
						wTmp[1][j][i] += rate * delta[maxPointer2nd][l][j]*unit[maxPointer2nd][l-1][i];

						wTmp[2][j][i] = (w[maxPointer1st][l][j][i] + w[maxPointer1stOther][l][j][i]) / 2.0;
						wTmp[2][j][i] += rate * delta[maxPointer1st][l][j]*unit[maxPointer1st][l-1][i];

						wTmp[3][j][i] = (w[maxPointer2nd][l][j][i] + w[maxPointer2ndOther][l][j][i]) / 2.0;
						wTmp[3][j][i] += rate * delta[maxPointer2nd][l][j]*unit[maxPointer2nd][l-1][i];

					}
				}

				w[0][l] = wTmp[0]; w[1][l] = wTmp[1]; w[2][l] = wTmp[2]; w[3][l] = wTmp[3];

			}
		}
	}

	public interface ReportFun { void action(int t, double err); }

	public void train21(int testSwitch, double[][] data, double[][] teach, double[][] testData, byte[] testLabel,
			int numberLoadData, int numberTrainData, int numberTestData, double learnRate, int numEpoch, int batchSize, ReportFun fun) {

		File wprt = new File("RecognitionRate" + numberLoadData + numberTrainData + numberTestData + numEpoch
				+ "testSwitch" + testSwitch + "b" + batchSize + ".csv");
		BufferedWriter fw = null;
		try {
			fw   = new BufferedWriter(new FileWriter(wprt, true));
		} catch (IOException e1) {
			e1.printStackTrace();
		}

		for (int epoch = 0; epoch < numEpoch ; epoch++) {
			mutationRateDominant = 0.0;
			mutationRateRecessive = 0.0;
			for (int k=0; k<numGamete; k++)
				errGamete[k] = 0.0;
			for (int t = 0; t < numberTrainData /batchSize ; t++) {
				for (int k=0; k < numGamete; k++){
					evalFuncValue[k] = 0.0;
					for (int i = 0; i < layerN; i++)
						for ( int j = 0; j < unitN[i]; j++){
							deltaCache[k][i][j] = 0.0;
							delta[k][i][j] = 0.0;
						}
				}

				for (int p = 0; p < batchSize; p++) {
					int idx = (t * batchSize + p) % numberTrainData;
					forward(data[idx]);
					backPropagate(data[idx], teach[idx]);
				}
				recombination2(testSwitch);
				err = 0.0;

				for (int k=0; k < numGamete; k++){
					evalFuncValue[k] = 0.0;
					for (int i = 0; i < layerN; i++)
						for ( int j = 0; j < unitN[i]; j++){
							deltaCache[k][i][j] = 0.0;
							delta[k][i][j] = 0.0;
						}
				}

				for (int p = 0; p < batchSize; p++) {
					int idx = (t * batchSize + p) % numberTrainData;
					forward(data[idx]);
					backPropagate(data[idx], teach[idx]);
				}
				update21(testSwitch, learnRate, batchSize);
			}

			if (fun != null) fun.action(epoch, err / numberTrainData / unitN[layerN-1]);

			for (int k=0; k < numGamete; k++){

				if ((0.99*errGameteOld[k] <= errGamete[k]) && (1.01*errGameteOld[k] >= errGamete[k])){
					if (epoch > 0) {
						mutationRateDominant = 0.9;
						mutationRateRecessive = 0.9;
					}
				}
				else {
					mutationRateDominant = 0.0;
					mutationRateRecessive = 0.0;
				}
				errGameteOld[k] = errGamete[k];
			}

			System.out.print("Epoch" + epoch + ": ");
			System.out.println("Sum of evaluation Function Values");
			for (int i=0; i < numGamete; i++) {
				System.out.print(i + ": ");
				System.out.printf(String.format("%.3f", evalFuncValue[i]));
				System.out.print(" ");
			}
			System.out.println();
			for (int i=0; i < numGamete; i++){
				System.out.print(i + ": ");
				System.out.printf(String.format("%.3f", evalFuncValue[i]/(double)numberTrainData));
				System.out.print(" ");
			}
			System.out.println();

			int n = numberTestData;
			int[] count = new int[numGamete];
			for (int i = 0; i < n; i++) {
				int label = testLabel[i];
				forward(testData[i]);
				for (int k = 0; k < numGamete; k++) {
					int result = getResult(k);
					if (result == label) count[k]++;
				}
			}
			try {
				fw.write("Epoch" + epoch + " Sum of evaluation Function Values");
				fw.newLine();
				fw.write(""+ n);
				fw.write(",");
				for (int k = 0; k < numGamete; k++){
					fw.write(""+ count[k]);
					fw.write(",");
				}
				fw.newLine();
				fw.write(""+ evalFuncValue[0]);
				fw.write(",");
				fw.write("" + evalFuncValue[1]);
				fw.write(",");
				fw.write("" + evalFuncValue[2]);
				fw.write(",");
				fw.write("" + evalFuncValue[3]);
				fw.write(",");
				fw.newLine();
				fw.write(""+ evalFuncValue[0]/(double)numberTrainData);
				fw.write(",");
				fw.write("" + evalFuncValue[1]/(double)numberTrainData);
				fw.write(",");
				fw.write("" + evalFuncValue[2]/(double)numberTrainData);
				fw.write(",");
				fw.write("" + evalFuncValue[3]/(double)numberTrainData);
				fw.write(",");
				fw.newLine();
				fw.write("Recognition Rate");
				fw.newLine();
				fw.write(""+ count[0] * 100.0 / n);
				fw.write("%,");
				fw.write("" + count[1] * 100.0 / n);
				fw.write("%,");
				fw.write("" + count[2] * 100.0 / n);
				fw.write("%,");
				fw.write("" + count[3] * 100.0 / n);
				fw.write("%,");
				fw.newLine();
			}
			catch (Exception e){
				System.out.println(e);
			}

		}	// t(Epoch) loop end
		System.out.println();
		try {
			fw.newLine();
		}
		catch (Exception e){
			System.out.println(e);
		}
		try {
			fw.flush();
		} catch (IOException e) {
			e.printStackTrace();
		}

		for (int i=0; i < numGamete; i++) evalFuncValue[i] = 0.0;
		err = 0.0;
		for (int i=0; i < numGamete; i++){
			errGamete[i] = 0.0;
		}
		for (int i= 0; i < numberTrainData ; i++) {
			forward(data[i]);
			backPropagate(data[i], teach[i]);
		}
		System.out.println("Sum of evaluation Function Values");
		for (int i=0; i < numGamete; i++){
			System.out.print(i + ": ");
			System.out.printf(String.format("%.3f", evalFuncValue[i]));
			System.out.print(" ");
		}
		System.out.println();
		for (int i=0; i < numGamete; i++) {
			System.out.print(i + ": ");
			System.out.printf(String.format("%.3f", evalFuncValue[i]/(double)numberTrainData));
			System.out.print(" ");
		}
		System.out.println();
		System.out.println();

	}

	public int getResult(int numberGamete) {
		double max = 0.0;
		int idx = -1;
		double[] out = unit[numberGamete][layerN - 1];
		for (int j = 0; j < out.length; j++) {
			if (out[j] > max) {
				max = out[j];
				idx = j;
			}
		}
		return idx;
	}
}
